/**
 * 事务处理
 * @type {async|exports|module.exports}
 */

var async = require('async');
var co = require('co');
var util = require('util');
var ObjectID = require('sails-mongo/node_modules/mongodb').ObjectID;
var sails = null;
module.exports = {
  init: function (sailsObject) {
    sails = sailsObject;
  },
  // START WATERLINE
  change: function (transactId, change, cb) {
    try {
      self.checkChange(change)
    } catch (e) {
      cb(e);
    }
    if (!change.time) {
      change.time = Date.now();
    }
    sails.models.transaction.native(function (err, collection) {
      var query = {
        $set: {
          state: 'pending'
        },
        $push: {changes: change}
      };
      collection.update({_id: new ObjectID(transactId)}, query, {upsert: false}, function (err, uTransaction) {
        cb(err)
      });
    });
  },
  checkChange: function (change) {
    return change;
  },
  start: function (transactionName, cb) {
    sails.models.transaction.create({
      name: transactionName || 'default',
      start_time: new Date(),
      state: 'start',
      changes: []
    }, function (err, tx) {
      console.log('ID:%s Name:%s 事务处理开始', tx.id, tx.name);
      cb(err, tx)
    })
  },
  finish: function (transactionId, cb) {
    sails.models.transaction.update({id: transactionId}, {state: 'finish', end_time: new Date()}, function () {
      console.log('ID:%s 事务处理完成', transactionId);
      console.log();
      cb();
    });

  },
  doRevertInsert: function (change, cb) {
    var model = change.coll;
    var docId = change.docId;
    if (!model || !docId) {
      return cb(util.format('doRevertInsert 出错 model %s 或者 id %s 为空', model, docId))
    }
    sails.models[model].destroy({id: docId}, function (err) {
      cb(err)
    })
  },
  doRevertUpdate: function (change, cb) {
    var model = change.coll;
    var docId = change.docId;
    var pre = change.pre;
    if (!model || !docId || !pre) {
      return cb(util.format('doRevertUpdate 出错 model %s 或者 id %s 为空 或者 pre %s为空', model, docId, pre))
    }
    sails.models[model].update({id: docId}, pre, function (err) {
      cb(err)
    })
  },
  rollback: function (transactionId, errMsg, cb) {
    sails.models.transaction.findOne({id: transactionId}, function (err, transact) {
      if (err || !transact) {
        cb(err || '事务不存在');
      }
      async.eachLimit(transact.changes, 1, function (change, each_cb) {
        switch (change.act) {
          case "insert" :
            self.doRevertInsert(change, each_cb)
            break;
          case "update":
            self.doRevertUpdate(change, each_cb)
            break;
          default :
            each_cb('未处理回滚');
            break;
        }
      }, function (err, results) {
        if (err) {
          sails.models.transaction.update({id: transactionId}, {
            state: 'cancelling',
            msg: err.toString(),
            end_time: new Date()
          }, function (rollback_err) {
            if (rollback_err) {
              console.log('ID:%s 设置事务回滚状态失败', transactionId, rollback_err);
            }
            console.log('ID:%s 事务回滚失败', transactionId, err);
            return cb(err);
          });
        }
        sails.models.transaction.update({id: transactionId}, {
          state: 'rollback',
          msg: errMsg,
          end_time: new Date()
        }, function (rollback_err) {
          if (!rollback_err) {
            console.error('ID:%s 事务回滚完成 ', transactionId);
            console.log();
          } else {
            console.log('ID:%s 设置事务回滚状态失败', transactionId, rollback_err);
          }
          cb(rollback_err);
        });
      });
    });

  },
  transactionManager: function (transactionName, task, callback) {
    var transactionId;
    async.waterfall([
      //事务开始
      function (cb) {
        self.start(transactionName, function (err, tx) {
          if (err || (!tx)) {
            cb(err || "can not begin transaction");
          } else {
            transactionId = tx.id;
            cb(null);
          }
        });
      },
      //方法执行
      function (cb) {
        task(transactionId, cb);
      }
    ], function (err, result) {
      if (err) {
        //事务回滚
        console.log('ID:%s 事务执行发生错误：', transactionId, err.stack);
        console.log('ID:%s 准备开始回滚', transactionId);
        self.rollback(transactionId, err, function (terr, tresult) {
          callback(terr, result);
        });
      } else {
        //事务完成
        self.finish(transactionId, function (terr, tresult) {
          callback(terr, result);
        });
      }
    });
  },

  transactionManagerPromise: function (transactionName, task) {
    return new Promise(function (resolve, reject) {
      self.transactionManagerCo(transactionName, task, function (err, result) {
        if (!err) {
          resolve(result);
        } else {
          reject(err);
        }
      })
    });
  },

  transactionManagerCo: function (transactionName, task, callback) {
    var transactionId;
    async.waterfall([
      //事务开始
      function (cb) {
        self.start(transactionName, function (err, tx) {
          if (err || (!tx)) {
            cb(err || "can not begin transaction");
          } else {
            transactionId = tx.id;
            cb(null);
          }
        });
      },
      //方法执行
      function (cb) {
        co(task(transactionId)).then(function (value) {
          console.log('ID:%s task result：', transactionId, value);
          //console.log(value);
          cb(null, value)
        }, function (err) {
          console.error('ID:%s task err：', transactionId, err);
          cb(err)
        });
      }
    ], function (err, result) {
      if (err) {
        //事务回滚
        console.error('ID:%s 事务执行发生错误：', transactionId, err);
        console.error('ID:%s 准备开始回滚', transactionId);
        self.rollback(transactionId, err, function (terr, tresult) {
          callback(terr, result);
        });
      } else {
        //事务完成
        self.finish(transactionId, function (terr, tresult) {
          console.log('ID:%s 事务执行完成', transactionId);
          callback(terr, result);
        });
      }
    });
  },

  txCreatePromise: function (Model, transactionId, values) {
    return new Promise(function (resolve, reject) {
      self.txCreate(Model, transactionId, values, function (err, result) {
        if (!err) {
          resolve(result);
        } else {
          reject(err);
        }
      })
    });
  },

  // 在事务中保存对象
  txCreate: function (Model, transactionId, values, CB) {
    // console.log('txCreate Model', typeof ( Model.create));
    async.waterfall([
        function (ascb) {
          Model.create(values, function (err, result) {
            if (!result) {
              ascb('txCreate 失败' + err);
            } else {
              ascb(err, result);
            }
          });
        },
        function (insertObject, ascb) {
          var change = {
            coll: Model.adapter.identity,
            act: 'insert',
            docId: insertObject.id,
            data: values,
            pre: null
          };
          self.change(transactionId, change, function (err) {
            ascb(err, insertObject)
          });
        }
      ],
      function (err, insertObject) {
        CB(err, insertObject);
      });
  },

  txUpdatePromise: function (Model, transactionId, query, values) {
    return new Promise(function (resolve, reject) {
      self.txUpdate(Model, transactionId, query, values, function (err, result) {
        if (!err) {
          resolve(result);
        } else {
          reject(err);
        }
      })
    });
  },

  // 在事务中保存对象
  txUpdate: function (Model, transactionId, query, values, CB) {
    async.waterfall([
        function (ascb) {
          Model.count(query, function (err, count) {
            if (count && count > 1) {
              err = '事务目前只支持单个文档的操作，在update操作中使用事务时，只能 update 单个文档。';
            }
            ascb(err);
          });
        },
        function (ascb) {
          Model.find(query, function (err, results) {
            if (err) return ascb(err);
            ascb(null, results[0]);
          });
        },
        function (pre, ascb) {
          // Update matching documents
          Model.update(query, values, function (err, results) {
            ascb(err, pre, results);
          });
        },
        function (pre, results, ascb) {
          var change = {
            coll: Model.adapter.identity,
            act: 'update',
            docId: pre.id,
            query: query,
            data: values,
            pre: pre
          };
          self.change(transactionId, change, function (err) {
            ascb(err, results)
          });
        }
      ],
      function (err, results) {
        CB(err, results);
      });
  },

  // 在事务中保存对象
  txNative: function (Model, transactionId, query, values, options, CB) {
    if (arguments.length === 5) {
      CB = options;
    }
    async.waterfall([
        function (ascb) {
          Model.native(function (err, collection) {
            collection.count(query, function (err, count) {
              if (count && count > 1) {
                err = '事务目前只支持单个文档的操作，在update操作中使用事务时，只能 update 单个文档。';
              }
              ascb(err);
            });
          });
        },
        function (ascb) {
          Model.native(function (err, collection) {
            collection.find(query).toArray(function (err, results) {
              ascb(err, results[0]);
            });
          });
        },
        function (pre, ascb) {
          // Update matching documents
          if (typeof(options) === 'function') {
            Model.native(function (err, collection) {
              collection.update(query, values, function (err, result) {
                ascb(err, pre, result);
              });
            });
          } else {
            Model.native(function (err, collection) {
              collection.update(query, values, options, function (err, result) {
                ascb(err, pre, result);
              });
            });
          }
        },
        function (pre, result, ascb) {
          var change = {
            coll: Model.adapter.identity,
            act: 'update',
            docId: pre._id.toString(),
            query: JSON.stringify(query),
            data: JSON.stringify(values),
            pre: pre
          };
          self.change(transactionId, change, function (err) {
            ascb(err, result)
          });
        }
      ],
      function (err, results) {
        CB(err, results);
      });
  }

};
var self = module.exports;